from __future__ import annotations

import asyncio
import hashlib
import logging
from abc import abstractmethod, ABC
from asyncio import Queue, Task
from datetime import timedelta
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Tuple, Optional, TypeVar, Type, cast
from typing import Callable, Awaitable

from aiohttp import ClientSession, TCPConnector
from aiohttp.web import Request
from arango.client import ArangoClient
from arango.database import StandardDatabase
from attr import define

from resotocore.analytics import AnalyticsEventSender
from resotocore.async_extensions import run_async
from resotocore.cli.cli import CLIService
from resotocore.cli.command import all_commands, alias_names
from resotocore.cli.model import CLI
from resotocore.config import ConfigHandler, ConfigOverride, NoConfigOverride
from resotocore.config.config_handler_service import ConfigHandlerService
from resotocore.config.core_config_handler import CoreConfigHandler
from resotocore.core_config import CoreConfig
from resotocore.db import SystemData
from resotocore.db.arangodb_extensions import ArangoHTTPClient
from resotocore.db.db_access import DbAccess
from resotocore.db.system_data_db import JwtSigningKeyHolder
from resotocore.graph_manager.graph_manager import GraphManager
from resotocore.infra_apps.local_runtime import LocalResotocoreAppRuntime
from resotocore.infra_apps.package_manager import PackageManager
from resotocore.infra_apps.runtime import Runtime
from resotocore.message_bus import MessageBus
from resotocore.metrics import timed
from resotocore.model.adjust_node import NoAdjust
from resotocore.model.db_updater import GraphMerger
from resotocore.model.model_handler import ModelHandler, ModelHandlerFromCodeAndDB
from resotocore.query.template_expander import TemplateExpander
from resotocore.query.template_expander_service import TemplateExpanderService
from resotocore.report import Inspector
from resotocore.report.inspector_service import InspectorFileService
from resotocore.service import Service
from resotocore.system_start import SystemInfo
from resotocore.task.scheduler import NoScheduler
from resotocore.task.subscribers import SubscriptionHandler, NoSubscriptionHandler
from resotocore.task.task_handler import TaskHandlerService
from resotocore.types import JsonElement
from resotocore.user import UserManagement
from resotocore.user.user_management import UserManagementService
from resotocore.util import Periodic
from resotocore.web.certificate_handler import CertificateHandler
from resotocore.worker_task_queue import WorkerTaskQueue

T = TypeVar("T")
log = logging.getLogger(__name__)


class ServiceNames:
    temp_dir = "temp_dir"
    cert_handler = "cert_handler"
    cli = "cli"
    config = "config"
    config_handler = "config_handler"
    config_override = "config_override"
    core_config_handler = "core_config_handler"
    db_access = "db_access"
    event_emitter_periodic = "event_emitter_periodic"
    event_sender = "event_sender"
    forked_tasks = "forked_tasks"
    graph_manager = "graph_manager"
    graph_merger = "graph_merger"
    http_session = "http_session"
    infra_apps_package_manager = "infra_apps_package_manager"
    infra_apps_runtime = "infra_apps_runtime"
    inspector = "inspector"
    jwt_signing_key_holder = "jwt_signing_key_holder"
    merge_outer_edges_handler = "merge_outer_edges_handler"
    message_bus = "message_bus"
    model_handler = "model_handler"
    scheduler = "scheduler"
    subscription_handler = "subscription_handler"
    system_data = "system_data"
    system_database = "system_database"
    system_info = "system_info"
    task_handler = "task_handler"
    template_expander = "template_expander"
    user_management = "user_management"
    worker_task_queue = "worker_task_queue"
    tenant_dependency_provider = "tenant_dependency_provider"


class Dependencies(Service):
    def __init__(self, **deps: Any) -> None:
        super().__init__()
        self.lookup: Dict[str, Any] = deps

    def add(self, name: str, service: T) -> "T":
        self.lookup[name] = service
        return service

    def extend(self, **deps: Any) -> "Dependencies":
        self.lookup = {**self.lookup, **deps}
        return self

    def get(self, name: str) -> Optional[Any]:
        return self.lookup.get(name)

    def all(self) -> Dict[str, Any]:
        return self.lookup

    @property
    def services(self) -> List[Service]:
        return [v for _, v in self.all().items() if isinstance(v, Service)]

    @property
    def tenant_dependency_provider(self) -> TenantDependencyProvider:
        return self.service(ServiceNames.tenant_dependency_provider, TenantDependencyProvider)  # type: ignore

    @property
    def config(self) -> CoreConfig:
        return self.service(ServiceNames.config, CoreConfig)

    @property
    def temp_dir(self) -> Path:
        return self.service(ServiceNames.temp_dir, Path)

    @property
    def message_bus(self) -> MessageBus:
        return self.service(ServiceNames.message_bus, MessageBus)

    @property
    def event_sender(self) -> AnalyticsEventSender:
        return self.service(ServiceNames.event_sender, AnalyticsEventSender)  # type: ignore

    @property
    def worker_task_queue(self) -> WorkerTaskQueue:
        return self.service(ServiceNames.worker_task_queue, WorkerTaskQueue)

    @property
    def jwt_signing_key_holder(self) -> JwtSigningKeyHolder:
        return self.service(ServiceNames.jwt_signing_key_holder, JwtSigningKeyHolder)  # type: ignore

    @property
    def system_info(self) -> SystemInfo:
        return self.service(ServiceNames.system_info, SystemInfo)

    @property
    def system_data(self) -> SystemData:
        return self.service(ServiceNames.system_data, SystemData)

    @property
    def forked_tasks(self) -> Queue[Tuple[Task[JsonElement], str]]:
        return self.service(ServiceNames.forked_tasks, Any)  # type:ignore

    @property
    def cert_handler(self) -> CertificateHandler:
        return self.service(ServiceNames.cert_handler, CertificateHandler)

    @property
    def http_session(self) -> ClientSession:
        session: Optional[ClientSession] = self.get(ServiceNames.http_session)
        if not session:
            connector = TCPConnector(limit=0, ssl=False, ttl_dns_cache=300)
            session = ClientSession(connector=connector)
            self.lookup[ServiceNames.http_session] = session
        return session

    def service(self, name: str, clazz: Type[T]) -> T:
        existing = self.get(name)
        if existing is None:
            raise KeyError(f"Service {name} not found")
        elif clazz is Any or isinstance(existing, clazz):
            return existing  # type: ignore
        else:
            raise ValueError(f"Service {name} is not of type {clazz}")

    def tenant_dependencies(self, **deps: Any) -> TenantDependencies:
        return TenantDependencies(nested=self, **deps)

    async def start(self) -> None:
        for service in self.services:
            await service.start()

    async def stop(self) -> None:
        if session := cast(Optional[ClientSession], self.get(ServiceNames.http_session)):
            await session.close()
        for service in reversed(self.services):
            await service.stop()


class TenantDependencies(Dependencies):
    def __init__(self, nested: Optional[Dependencies] = None, **deps: Any) -> None:
        super().__init__(**deps)
        self.nested = nested or Dependencies()

    def get(self, name: str) -> Optional[Any]:
        if (defined := self.lookup.get(name)) is not None:
            return defined
        else:
            return self.nested.get(name)

    @property
    def subscription_handler(self) -> SubscriptionHandler:
        return self.service(ServiceNames.subscription_handler, SubscriptionHandler)  # type: ignore

    @property
    def config_handler(self) -> ConfigHandler:
        return self.service(ServiceNames.config_handler, ConfigHandler)  # type: ignore

    @property
    def db_access(self) -> DbAccess:
        return self.service(ServiceNames.db_access, DbAccess)

    @property
    def model_handler(self) -> ModelHandler:
        return self.service(ServiceNames.model_handler, ModelHandler)  # type: ignore

    @property
    def inspector(self) -> Inspector:
        return self.service(ServiceNames.inspector, Inspector)  # type: ignore

    @property
    def infra_apps_runtime(self) -> Runtime:
        return self.service(ServiceNames.infra_apps_runtime, Runtime)  # type: ignore

    @property
    def infra_apps_package_manager(self) -> PackageManager:
        return self.service(ServiceNames.infra_apps_package_manager, PackageManager)

    @property
    def user_management(self) -> UserManagement:
        return self.service(ServiceNames.user_management, UserManagement)  # type: ignore

    @property
    def graph_manager(self) -> GraphManager:
        return self.service(ServiceNames.graph_manager, GraphManager)

    @property
    def graph_merger(self) -> GraphMerger:
        return self.service(ServiceNames.graph_merger, GraphMerger)

    @property
    def config_override(self) -> ConfigOverride:
        return self.service(ServiceNames.config_override, ConfigOverride)  # type: ignore

    @property
    def template_expander(self) -> TemplateExpander:
        return self.service(ServiceNames.template_expander, TemplateExpander)  # type: ignore

    @property
    def task_handler(self) -> TaskHandlerService:
        return self.service(ServiceNames.task_handler, TaskHandlerService)

    @property
    def cli(self) -> CLI:
        return self.service(ServiceNames.cli, CLI)  # type: ignore


class TenantDependencyProvider(Service, ABC):
    @abstractmethod
    async def dependencies(self, request: Request) -> TenantDependencies:
        pass


class DirectTenantDependencyProvider(TenantDependencyProvider):
    def __init__(self, dependencies: TenantDependencies) -> None:
        super().__init__()
        self._dependencies = dependencies

    async def dependencies(self, _: Request) -> TenantDependencies:
        return self._dependencies


class TenantDependencyCache(Service):
    def __init__(
        self, ttl: timedelta, check_frequency: timedelta, time_fn: Optional[Callable[[], float]] = None
    ) -> None:
        super().__init__()
        self._ttl = ttl.total_seconds()
        self._cache: Dict[str, Tuple[float, TenantDependencies]] = {}
        self._locks: Dict[str, asyncio.Lock] = {}
        self._locks_lock: Optional[asyncio.Lock] = None
        self.periodic = Periodic("tenant_dependency_cache", self._expire, check_frequency)
        self.time_fn = time_fn

    async def start(self) -> Any:
        self._locks_lock = asyncio.Lock()
        await self.periodic.start()

    async def stop(self) -> None:
        await self.periodic.stop()
        for _, value in self._cache.values():
            await value.stop()

    def _time(self) -> float:
        return (self.time_fn or asyncio.get_event_loop().time)()

    async def _lock_for(self, key: str) -> asyncio.Lock:
        assert self._locks_lock is not None, "TenantDependencyCache not started"
        async with self._locks_lock:
            if lock := self._locks.get(key):
                return lock
            else:
                lock = asyncio.Lock()
                self._locks[key] = lock
                return lock

    async def _expire(self) -> None:
        now = self._time()
        to_delete = []
        for key, (timestamp, value) in list(self._cache.items()):
            lock = await self._lock_for(key)
            async with lock:
                if now - timestamp > self._ttl:
                    to_delete.append((key, value))
                    self._cache.pop(key, None)
                    # content lock is not removed on purpose.
        for key, value in to_delete:
            log.info(f"Stop tenant dependencies for {key}")
            await value.stop()

    async def get(self, key: str, if_empty: Callable[[], Awaitable[TenantDependencies]]) -> TenantDependencies:
        now = self._time()
        lck = await self._lock_for(key)
        async with lck:
            if result := self._cache.get(key):
                _, value = result
            else:
                log.info(f"Create and start new tenant dependencies for {key}")
                value = await if_empty()
                await value.start()
            self._cache[key] = (now, value)
            return value


@define
class GraphDbAccess:
    server: str
    database: str
    username: str
    password: str

    def is_valid(self) -> bool:
        return bool(self.server and self.database and self.username)

    def hash(self) -> str:
        sha256 = hashlib.sha256()
        for key in ["server", "database", "username", "password"]:
            value = getattr(self, key)
            sha256.update(value.encode("utf-8"))
        return sha256.hexdigest()


class FromRequestTenantDependencyProvider(TenantDependencyProvider):
    def __init__(self, dependencies: Dependencies) -> None:
        super().__init__()
        self._dependencies = dependencies
        self._cache = TenantDependencyCache(ttl=timedelta(minutes=5), check_frequency=timedelta(seconds=30))

    async def start(self) -> None:
        await self._cache.start()

    async def stop(self) -> None:
        await self._cache.stop()

    @timed("tenant_dependency_provider", "dependencies")
    async def dependencies(self, request: Request) -> TenantDependencies:
        db_access = GraphDbAccess(
            request.headers.get("FixGraphDbServer", ""),
            request.headers.get("FixGraphDbDatabase", ""),
            request.headers.get("FixGraphDbUsername", ""),
            request.headers.get("FixGraphDbPassword", ""),
        )
        if not db_access.is_valid():
            raise ValueError("Invalid graph db access data provided for multi tenant requests!")
        key = db_access.hash()
        return await self._cache.get(key, partial(self.create_tenant_dependencies, key, db_access))

    async def create_tenant_dependencies(self, tenant_hash: str, access: GraphDbAccess) -> TenantDependencies:
        dp = self._dependencies
        config = dp.config
        args = dp.config.args
        message_bus = dp.message_bus
        event_sender = dp.event_sender

        def standard_database() -> StandardDatabase:
            http_client = ArangoHTTPClient(args.graphdb_request_timeout, verify=dp.config.run.verify)
            client = ArangoClient(hosts=access.server, http_client=http_client)
            return client.db(name=access.database, username=access.username, password=access.password)

        deps = self._dependencies.tenant_dependencies(tenant_hash=tenant_hash, access=access)
        # direct db access
        sdb = deps.add(ServiceNames.system_database, await run_async(standard_database))
        db = deps.add(ServiceNames.db_access, DbAccess(sdb, dp.event_sender, NoAdjust(), config))
        # no scheduler required in multi-tenant mode
        scheduler = deps.add(ServiceNames.scheduler, NoScheduler())
        # all tenants use the same model (derived from code)
        model = deps.add(ServiceNames.model_handler, ModelHandlerFromCodeAndDB(db, config.runtime.plantuml_server))
        worker_task_queue = deps.add(ServiceNames.worker_task_queue, WorkerTaskQueue())
        config_override_service = deps.add(ServiceNames.config_override, NoConfigOverride())
        config_handler = deps.add(
            ServiceNames.config_handler,
            ConfigHandlerService(
                db.config_entity_db,
                db.config_validation_entity_db,
                db.configs_model_db,
                worker_task_queue,
                message_bus,
                event_sender,
                config,
                config_override_service,
            ),
        )
        deps.add(ServiceNames.user_management, UserManagementService(db, config_handler, event_sender))
        default_env = {"graph": config.cli.default_graph, "section": config.cli.default_section}
        cli = deps.add(ServiceNames.cli, CLIService(deps, all_commands(deps), default_env, alias_names()))
        deps.add(ServiceNames.template_expander, TemplateExpanderService(db.template_entity_db, cli))
        inspector = deps.add(ServiceNames.inspector, InspectorFileService(cli))
        subscriptions = deps.add(ServiceNames.subscription_handler, NoSubscriptionHandler())
        core_config_handler = deps.add(
            ServiceNames.core_config_handler,
            CoreConfigHandler(config, message_bus, worker_task_queue, config_handler, event_sender, inspector),
        )
        deps.add(ServiceNames.infra_apps_runtime, LocalResotocoreAppRuntime(cli))
        deps.add(
            ServiceNames.infra_apps_package_manager,
            PackageManager(
                db.package_entity_db, config_handler, cli.register_infra_app_alias, cli.unregister_infra_app_alias
            ),
        )
        graph_merger = deps.add(ServiceNames.graph_merger, GraphMerger(model, event_sender, config, message_bus))
        task_handler = deps.add(
            ServiceNames.task_handler,
            TaskHandlerService(
                db.running_task_db,
                db.job_db,
                message_bus,
                event_sender,
                subscriptions,
                graph_merger,
                scheduler,
                cli,
                config,
            ),
        )
        deps.add(ServiceNames.graph_manager, GraphManager(db, config, core_config_handler, task_handler))
        return deps



import asyncio
import logging
from abc import abstractmethod
from collections import defaultdict
from functools import lru_cache
from typing import Optional, List, Dict, Tuple, Callable, AsyncIterator, cast

from aiostream import stream
from attr import define

from resotocore.analytics import CoreEvent
from resotocore.cli.model import CLIContext, CLI
from resotocore.config import ConfigEntity, ConfigHandler
from resotocore.db.model import QueryModel
from resotocore.error import NotFoundError
from resotocore.ids import ConfigId, GraphName, NodeId
from resotocore.model.model import Model
from resotocore.model.resolve_in_graph import NodePath
from resotocore.query.model import Query, P, Term
from resotocore.report import (
    Inspector,
    ReportCheck,
    Benchmark,
    BenchmarkResult,
    CheckCollection,
    CheckCollectionResult,
    CheckResult,
    CheckConfigPrefix,
    BenchmarkConfigPrefix,
    CheckConfigRoot,
    ResotoReportValues,
    BenchmarkConfigRoot,
    ResotoReportBenchmark,
    ResotoReportCheck,
    ReportSeverity,
    ReportSeverityPriority,
)
from resotocore.report.report_config import ReportCheckCollectionConfig, BenchmarkConfig
from resotocore.service import Service
from resotocore.types import Json
from resotocore.util import value_in_path, uuid_str, value_in_path_get
from resotolib.json_bender import Bender, S, bend

log = logging.getLogger(__name__)

SingleCheckResult = Dict[str, List[Json]]


def benchmark_id(name: str) -> ConfigId:
    return ConfigId(BenchmarkConfigPrefix + name)


def check_id(name: str) -> ConfigId:
    return ConfigId(CheckConfigPrefix + name)


@define
class CheckContext:
    accounts: Optional[List[str]] = None
    severity: Optional[ReportSeverity] = None
    only_failed: bool = False
    parallel_checks: int = 10

    def severities_including(self, severity: ReportSeverity) -> List[ReportSeverity]:
        return [s for s in ReportSeverity if self.includes_severity(severity)]

    def includes_severity(self, severity: ReportSeverity) -> bool:
        if self.severity is None:
            return True
        else:
            return ReportSeverityPriority[self.severity] <= ReportSeverityPriority[severity]


# This defines the subset of the data provided for every resource
ReportResourceData: Dict[str, Bender] = {
    "node_id": S("id"),
    "id": S("reported", "id"),
    "name": S("reported", "name"),
    "kind": S("reported", "kind"),
    "tags": S("reported", "tags"),
    "ctime": S("reported", "ctime"),
    "atime": S("reported", "atime"),
    "mtime": S("reported", "mtime"),
    "cloud": S("ancestors", "cloud", "reported", "name"),
    "account": S("ancestors", "account", "reported", "name"),
    "region": S("ancestors", "region", "reported", "name"),
    "zone": S("ancestors", "zone", "reported", "name"),
}


class InspectorService(Inspector, Service):
    def __init__(self, cli: CLI) -> None:
        super().__init__()
        self.db_access = cli.dependencies.db_access
        self.cli = cli
        self.template_expander = cli.dependencies.template_expander
        self.model_handler = cli.dependencies.model_handler
        self.event_sender = cli.dependencies.event_sender

    @abstractmethod
    async def _report_values(self) -> Json:
        pass

    @abstractmethod
    async def _check_ids(self) -> List[ConfigId]:
        pass

    @abstractmethod
    async def _checks(self, cfg_id: ConfigId) -> List[ReportCheck]:
        pass

    @abstractmethod
    async def _benchmark(self, cfg_id: ConfigId) -> Benchmark:
        pass

    async def benchmark(self, name: str) -> Optional[Benchmark]:
        try:
            return await self._benchmark(benchmark_id(name))
        except ValueError:
            return None

    async def list_checks(
        self,
        *,
        provider: Optional[str] = None,
        service: Optional[str] = None,
        category: Optional[str] = None,
        kind: Optional[str] = None,
        check_ids: Optional[List[str]] = None,
        context: Optional[CheckContext] = None,
    ) -> List[ReportCheck]:
        def inspection_matches(inspection: ReportCheck) -> bool:
            return (
                (provider is None or provider == inspection.provider)
                and (service is None or service == inspection.service)
                and (category is None or category in inspection.categories)
                and (kind is None or kind == inspection.result_kind)
                and (check_ids is None or inspection.id in check_ids)
                and (context is None or context.includes_severity(inspection.severity))
            )

        return await self.filter_checks(inspection_matches)

    async def load_benchmarks(
        self,
        graph: GraphName,
        benchmark_names: List[str],
        *,
        accounts: Optional[List[str]] = None,
        severity: Optional[ReportSeverity] = None,
        only_failing: bool = False,
    ) -> Dict[str, BenchmarkResult]:
        context = CheckContext(accounts=accounts, severity=severity, only_failed=only_failing)
        # create query
        term: Term = P("benchmark").is_in(benchmark_names)
        if severity:
            term = term & P("severity").is_in([s.value for s in context.severities_including(severity)])
        term = P.context("security.issues[]", term)
        if accounts:
            term = term & P("ancestors.account.reported.id").is_in(accounts)
        term = term & P("security.has_issues").eq(True)
        model = QueryModel(Query.by(term), await self.model_handler.load_model(graph))

        # collect all checks
        benchmarks = {name: await self._benchmark(benchmark_id(name)) for name in benchmark_names}
        check_ids = {check for b in benchmarks.values() for check in b.nested_checks()}
        checks = await self.list_checks(check_ids=list(check_ids), context=context)
        check_lookup = {check.id: check for check in checks}

        # perform query, map resources and create lookup map
        check_results: Dict[str, SingleCheckResult] = defaultdict(lambda: defaultdict(list))
        async with await self.db_access.get_graph_db(graph).search_list(model) as cursor:
            async for entry in cursor:
                if account_id := value_in_path(entry, NodePath.ancestor_account_id):
                    mapped = bend(ReportResourceData, entry)
                    for issue in value_in_path_get(entry, NodePath.security_issues, cast(List[Json], [])):
                        if check := issue.get("check"):
                            check_results[check][account_id].append(mapped)
        return {
            name: self.__to_result(benchmark, check_lookup, check_results, context)
            for name, benchmark in benchmarks.items()
        }

    async def perform_benchmarks(
        self,
        graph: GraphName,
        benchmark_names: List[str],
        *,
        accounts: Optional[List[str]] = None,
        severity: Optional[ReportSeverity] = None,
        only_failing: bool = False,
        sync_security_section: bool = False,
        report_run_id: Optional[str] = None,
    ) -> Dict[str, BenchmarkResult]:
        context = CheckContext(accounts=accounts, severity=severity, only_failed=only_failing)
        benchmarks = {name: await self._benchmark(benchmark_id(name)) for name in benchmark_names}
        # collect all checks
        check_ids = {check for b in benchmarks.values() for check in b.nested_checks()}
        checks = await self.list_checks(check_ids=list(check_ids), context=context)
        check_lookup = {check.id: check for check in checks}
        # create benchmark results
        results = await self.__perform_checks(graph, checks, context)
        result = {
            name: self.__to_result(benchmark, check_lookup, results, context) for name, benchmark in benchmarks.items()
        }
        if sync_security_section:
            model = await self.model_handler.load_model(graph)
            # In case no run_id is provided, we invent a report run id here.
            run_id = report_run_id or uuid_str()
            await self.db_access.get_graph_db(graph).update_security_section(
                run_id, self.__benchmarks_to_security_iterator(result), model
            )
        return result

    async def perform_checks(
        self,
        graph: GraphName,
        *,
        provider: Optional[str] = None,
        service: Optional[str] = None,
        category: Optional[str] = None,
        kind: Optional[str] = None,
        check_ids: Optional[List[str]] = None,
        accounts: Optional[List[str]] = None,
        severity: Optional[ReportSeverity] = None,
        only_failing: bool = False,
    ) -> BenchmarkResult:
        context = CheckContext(accounts=accounts, severity=severity, only_failed=only_failing)
        checks = await self.list_checks(
            provider=provider, service=service, category=category, kind=kind, check_ids=check_ids, context=context
        )
        provider_name = f"{provider}_" if provider else ""
        service_name = f"{service}_" if service else ""
        category_name = f"{category}_" if category else ""
        kind_name = f"{kind}_" if kind else ""
        check_id_name = f"{check_ids[0]}_" if check_ids else ""
        title = f"{provider_name}{service_name}{category_name}{kind_name}{check_id_name}benchmark"
        benchmark = Benchmark(
            id=title,
            title=title,
            description="On demand benchmark",
            documentation="On demand benchmark",
            framework="resoto",
            version="1.0",
            checks=[c.id for c in checks],
            children=[],
        )

        if context.accounts is None:
            context.accounts = await self.__list_accounts(benchmark, graph)

        checks_to_perform = await self.list_checks(check_ids=benchmark.nested_checks(), context=context)
        check_by_id = {c.id: c for c in checks_to_perform}
        results = await self.__perform_checks(graph, checks_to_perform, context)
        await self.event_sender.core_event(CoreEvent.BenchmarkPerformed, {"benchmark": benchmark.id})
        return self.__to_result(benchmark, check_by_id, results, context)

    async def filter_checks(self, report_filter: Optional[Callable[[ReportCheck], bool]] = None) -> List[ReportCheck]:
        cfg_ids = await self._check_ids()
        list_of_lists = await asyncio.gather(*[self._checks(cfg_id) for cfg_id in cfg_ids])
        return [
            check for entries in list_of_lists for check in entries if report_filter is None or report_filter(check)
        ]

    async def list_failing_resources(
        self, graph: GraphName, check_uid: str, account_ids: Optional[List[str]] = None
    ) -> AsyncIterator[Json]:
        # create context
        context = CheckContext(accounts=account_ids)
        # get check
        checks = await self.list_checks(check_ids=[check_uid], context=context)
        if not checks:
            raise NotFoundError(f"Check {check_uid} not found")
        model = await self.model_handler.load_model(graph)
        inspection = checks[0]
        # load configuration
        cfg = await self._report_values()
        return await self.__list_failing_resources(graph, model, inspection, cfg, context)

    async def __list_failing_resources(
        self, graph: GraphName, model: Model, inspection: ReportCheck, config: Json, context: CheckContext
    ) -> AsyncIterator[Json]:
        # final environment: defaults are coming from the check and are eventually overriden in the config
        env = inspection.environment(config)
        account_id_prop = "ancestors.account.reported.id"
        # if the result kind is an account, we need to use the id directly instead of walking the graph
        if (result_kind := model.get(inspection.result_kind)) and "account" in result_kind.kind_hierarchy():
            account_id_prop = "reported.id"

        async def perform_search(search: str) -> AsyncIterator[Json]:
            # parse query
            query = await self.template_expander.parse_query(search, on_section="reported", **env)
            # filter only relevant accounts if provided
            if context.accounts:
                query = Query.by(P.single(account_id_prop).is_in(context.accounts)).combine(query)
            async with await self.db_access.get_graph_db(graph).search_list(QueryModel(query, model)) as ctx:
                async for result in ctx:
                    yield result

        async def perform_cmd(cmd: str) -> AsyncIterator[Json]:
            # filter only relevant accounts if provided
            if context.accounts:
                account_list = ",".join(f'"{a}"' for a in context.accounts)
                cmd = f"search /{account_id_prop} in [{account_list}] | " + cmd
            cli_result = await self.cli.execute_cli_command(cmd, stream.list, CLIContext(env=env))
            for result in cli_result[0]:
                yield result

        async def empty() -> AsyncIterator[Json]:
            if False:  # pylint: disable=using-constant-test
                yield {}  # noqa

        if resoto_search := inspection.detect.get("resoto"):
            return perform_search(resoto_search)
        elif resoto_cmd := inspection.detect.get("resoto_cmd"):
            return perform_cmd(resoto_cmd)
        else:
            return empty()

    def __to_result(
        self,
        benchmark: Benchmark,
        check_by_id: Dict[str, ReportCheck],
        results: Dict[str, SingleCheckResult],
        context: CheckContext,
    ) -> BenchmarkResult:
        def to_result(cc: CheckCollection) -> CheckCollectionResult:
            check_results = []
            for cid in cc.checks or []:
                if (check := check_by_id.get(cid)) is not None:
                    result = results.get(cid, {})
                    count_by_account = {uid: len(failed) for uid, failed in result.items()}
                    check_results.append(CheckResult(check, count_by_account, result))
            children = [to_result(c) for c in cc.children or []]
            return CheckCollectionResult(
                cc.title, cc.description, documentation=cc.documentation, checks=check_results, children=children
            )

        top = to_result(benchmark).filter_result(context.only_failed)
        return BenchmarkResult(
            benchmark.title,
            benchmark.description,
            benchmark.framework,
            benchmark.version,
            documentation=benchmark.documentation,
            checks=top.checks,
            children=top.children,
            accounts=context.accounts,
            only_failed=context.only_failed,
            severity=context.severity,
            id=benchmark.id,
        )

    async def __perform_checks(
        self, graph: GraphName, checks: List[ReportCheck], context: CheckContext
    ) -> Dict[str, SingleCheckResult]:
        # load model
        model = await self.model_handler.load_model(graph)
        # load configuration
        cfg = await self._report_values()

        async def perform_single(check: ReportCheck) -> Tuple[str, SingleCheckResult]:
            return check.id, await self.__perform_check(graph, model, check, cfg, context)

        async with stream.map(
            stream.iterate(checks), perform_single, ordered=False, task_limit=context.parallel_checks
        ).stream() as streamer:
            return {key: value async for key, value in streamer}

    async def __perform_check(
        self, graph: GraphName, model: Model, inspection: ReportCheck, config: Json, context: CheckContext
    ) -> SingleCheckResult:
        resources_by_account = defaultdict(list)
        # if the result kind is an account, we need to use the id directly instead of walking the graph
        is_account = (rk := model.get(inspection.result_kind)) and "account" in rk.kind_hierarchy()
        account_id_path = NodePath.reported_id if is_account else NodePath.ancestor_account_id
        async for resource in await self.__list_failing_resources(graph, model, inspection, config, context):
            account_id = value_in_path(resource, account_id_path)
            if account_id:
                resources_by_account[account_id].append(bend(ReportResourceData, resource))
        return resources_by_account

    async def __list_accounts(self, benchmark: Benchmark, graph: GraphName) -> List[str]:
        model = await self.model_handler.load_model(graph)
        gdb = self.db_access.get_graph_db(graph)
        query = Query.by("account")
        if benchmark.clouds:
            query = query.combine(Query.by(P.single("ancestors.cloud.reported.id").is_in(benchmark.clouds)))
        async with await gdb.search_list(QueryModel(query, model)) as crs:
            ids = [value_in_path(a, NodePath.reported_id) async for a in crs]
            return [aid for aid in ids if aid is not None]

    async def validate_benchmark_config(self, cfg_id: ConfigId, json: Json) -> Optional[Json]:
        try:
            benchmark = BenchmarkConfig.from_config(ConfigEntity(ResotoReportBenchmark, json))
            bid = cfg_id.rsplit(".", 1)[-1]
            if benchmark.id != bid:
                return {"error": f"Benchmark id should be {bid} (same as the config name). Got {benchmark.id}"}
            all_checks = {c.id for c in await self.filter_checks()}
            missing = []
            for check in benchmark.nested_checks():
                if check not in all_checks:
                    missing.append(check)
            if missing:
                return {"error": f"Following checks are defined in the benchmark but do not exist: {missing}"}
            else:
                return None
        except Exception as e:
            return {"error": f"Can not digest benchmark: {e}"}

    async def validate_check_collection_config(self, json: Json) -> Optional[Json]:
        try:
            errors = []
            for check in ReportCheckCollectionConfig.from_config(ConfigEntity(ResotoReportCheck, json)):
                try:
                    env = check.default_values or {}
                    if search := check.detect.get("resoto"):
                        await self.template_expander.parse_query(search, on_section="reported", **env)
                    elif cmd := check.detect.get("resoto_cmd"):
                        await self.cli.evaluate_cli_command(cmd, CLIContext(env=env))
                    elif check.detect.get("manual"):
                        pass
                    else:
                        errors.append(f"Check {check.id} neither has a resoto, resoto_cmd or manual defined")
                except Exception as e:
                    errors.append(f"Check {check.id} is invalid: {e}")
            if errors:
                return {"error": f"Can not validate check collection: {errors}"}
            else:
                return None

        except Exception as e:
            return {"error": f"Can not digest check collection: {e}"}

    def __benchmarks_to_security_iterator(
        self, results: Dict[str, BenchmarkResult]
    ) -> AsyncIterator[Tuple[NodeId, Json]]:
        # Create a mapping from node_id to all check results that contain this node
        node_result: Dict[str, List[Tuple[BenchmarkResult, CheckResult]]] = defaultdict(list)

        def walk_collection(collection: CheckCollectionResult, parent: BenchmarkResult) -> None:
            for check in collection.checks:
                for resources in check.resources_failing_by_account.values():
                    for resource in resources:
                        node_result[resource["node_id"]].append((parent, check))
            for child in collection.children:
                walk_collection(child, parent)

        for result in results.values():
            walk_collection(result, result)

        async def iterate_nodes() -> AsyncIterator[Tuple[NodeId, Json]]:
            for node_id, contexts in node_result.items():
                issues = [
                    dict(benchmark=bench.id, check=check.check.id, severity=check.check.severity.name)
                    for bench, check in contexts
                ]
                yield NodeId(node_id), dict(issues=issues)

        return iterate_nodes()


@lru_cache(maxsize=1)
def benchmarks_from_file() -> Dict[ConfigId, Benchmark]:
    result = {}
    for name, js in BenchmarkConfig.from_files().items():
        cid = benchmark_id(name)
        benchmark = BenchmarkConfig.from_config(ConfigEntity(cid, {BenchmarkConfigRoot: js}))
        result[cid] = benchmark
    return result


@lru_cache(maxsize=1)
def checks_from_file() -> Dict[ConfigId, List[ReportCheck]]:
    result = {}
    for name, js in ReportCheckCollectionConfig.from_files().items():
        cid = check_id(name)
        result[cid] = ReportCheckCollectionConfig.from_config(ConfigEntity(cid, {CheckConfigRoot: js}))
    return result


class InspectorFileService(InspectorService):
    async def _report_values(self) -> Json:
        return {}  # default values

    async def _check_ids(self) -> List[ConfigId]:
        return list(checks_from_file().keys())

    async def _checks(self, cfg_id: ConfigId) -> List[ReportCheck]:
        return checks_from_file().get(cfg_id, [])

    async def _benchmark(self, cfg_id: ConfigId) -> Benchmark:
        return benchmarks_from_file()[cfg_id]

    async def list_benchmarks(self) -> List[Benchmark]:
        return list(benchmarks_from_file().values())

    @staticmethod
    def on_startup() -> None:
        # make sure benchmarks and checks are loaded
        benchmarks_from_file()
        checks_from_file()


class InspectorConfigService(InspectorService):
    def __init__(self, cli: CLI) -> None:
        super().__init__(cli)
        self.config_handler: ConfigHandler = cli.dependencies.config_handler

    async def start(self) -> None:
        if not self.cli.dependencies.config.multi_tenant_setup:
            # TODO: we need a migration path for checks added in existing configs
            config_ids = {i async for i in self.config_handler.list_config_ids()}
            overwrite = False  # Only here to simplify development. True until we reach a stable version.
            for name, js in BenchmarkConfig.from_files().items():
                if overwrite or benchmark_id(name) not in config_ids:
                    cid = benchmark_id(name)
                    log.info(f"Creating benchmark config {cid}")
                    await self.config_handler.put_config(ConfigEntity(cid, {BenchmarkConfigRoot: js}), validate=False)
            for name, js in ReportCheckCollectionConfig.from_files().items():
                if overwrite or check_id(name) not in config_ids:
                    cid = check_id(name)
                    log.info(f"Creating check collection config {cid}")
                    await self.config_handler.put_config(ConfigEntity(cid, {CheckConfigRoot: js}), validate=False)

    async def _report_values(self) -> Json:
        cfg_entity = await self.config_handler.get_config(ResotoReportValues)
        return cfg_entity.config if cfg_entity else {}

    async def _check_ids(self) -> List[ConfigId]:
        return [i async for i in self.config_handler.list_config_ids() if i.startswith(CheckConfigPrefix)]

    async def _checks(self, cfg_id: ConfigId) -> List[ReportCheck]:
        config = await self.config_handler.get_config(cfg_id)
        if config is not None and CheckConfigRoot in config.config:
            return ReportCheckCollectionConfig.from_config(config)
        else:
            return []

    async def _benchmark(self, cfg_id: ConfigId) -> Benchmark:
        cfg = await self.config_handler.get_config(cfg_id)
        if cfg is None or BenchmarkConfigRoot not in cfg.config:
            raise ValueError(f"Unknown benchmark: {cfg_id}")
        return BenchmarkConfig.from_config(cfg)

    async def list_benchmarks(self) -> List[Benchmark]:
        return [
            await self._benchmark(i)
            async for i in self.config_handler.list_config_ids()
            if i.startswith(BenchmarkConfigPrefix)
        ]



from __future__ import annotations

import logging
from datetime import timedelta
from pathlib import Path
from ssl import SSLContext, create_default_context, Purpose
from tempfile import TemporaryDirectory
from typing import Tuple, Optional, List, Union, Dict

from arango.database import StandardDatabase
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509 import Certificate, CertificateSigningRequest

from resotocore.core_config import CoreConfig, CertificateConfig
from resotocore.service import Service
from resotocore.types import Json
from resotocore.util import Periodic
from resotolib.core.ca import TLSData
from resotolib.utils import get_local_ip_addresses, get_local_hostnames
from resotolib.x509 import (
    bootstrap_ca,
    gen_rsa_key,
    gen_csr,
    sign_csr,
    cert_to_bytes,
    key_to_bytes,
    load_key_from_bytes,
    load_cert_from_bytes,
    cert_fingerprint,
    load_csr_from_bytes,
    write_key_to_file,
    write_cert_to_file,
    load_key_from_file,
    load_cert_from_file,
    write_ca_bundle,
)

log = logging.getLogger(__name__)


class CertificateHandler(Service):
    def __init__(
        self, config: CoreConfig, ca_cert: Certificate, host_key: RSAPrivateKey, host_cert: Certificate, temp_dir: Path
    ) -> None:
        super().__init__()
        self.config = config
        self._ca_cert = ca_cert
        self._ca_cert_bytes = cert_to_bytes(ca_cert)
        self._ca_cert_fingerprint = cert_fingerprint(ca_cert)
        self._ca_bundle = temp_dir / "ca-bundle.crt"
        self._host_key = host_key
        self._host_cert = host_cert
        self.__recreate_ca_file()  # write the CA bundle to the temp dir
        self._ca_cert_recreate = Periodic("recreate ca bundle file", self.__recreate_ca_file, timedelta(hours=1))
        self._host_context = self._create_host_context(config, self._host_cert, self._host_key)
        self._client_context = self.__create_client_context(config, ca_cert)

    async def start(self) -> None:
        await self._ca_cert_recreate.start()

    async def stop(self) -> None:
        await self._ca_cert_recreate.stop()

    def __recreate_ca_file(self) -> None:
        write_ca_bundle(self._ca_cert, str(self._ca_bundle))

    @property
    def ca_cert(self) -> Certificate:
        return self._ca_cert

    @property
    def ca_bundle(self) -> Path:
        return self._ca_bundle

    @property
    def authority_certificate(self) -> Tuple[bytes, str]:
        return self._ca_cert_bytes, self._ca_cert_fingerprint

    @property
    def host_certificate(self) -> Tuple[RSAPrivateKey, Certificate]:
        return self._host_key, self._host_cert

    def create_key_and_cert(
        self, common_name: str, dns_names: List[str], ip_addresses: List[str], days_valid: int
    ) -> Tuple[RSAPrivateKey, Certificate]:
        raise NotImplementedError("Signing is not implemented!")

    def sign(
        self,
        csr_or_bytes: Union[CertificateSigningRequest, bytes],
        days_valid: int = 365,
        server_auth: bool = True,
        client_auth: bool = True,
        key_usage: Optional[Dict[str, bool]] = None,
    ) -> Tuple[Certificate, str]:
        raise NotImplementedError("Signing is not implemented!")

    @property
    def host_context(self) -> Optional[SSLContext]:
        return self._host_context

    @property
    def client_context(self) -> SSLContext:
        return self._client_context

    @staticmethod
    def _create_host_context(config: CoreConfig, host_cert: Certificate, host_key: RSAPrivateKey) -> SSLContext:
        args = config.args
        # noinspection PyTypeChecker
        ctx = create_default_context(purpose=Purpose.CLIENT_AUTH)
        if config.args.cert:
            log.info("Using TLS certificate from command line.")
            # Use the certificate provided via cmd line flags
            ctx.load_cert_chain(args.cert, args.cert_key, args.cert_key_pass)
        else:
            log.info("Using TLS certificate from data store.")
            # ssl library wants to load cert/key from file: put it into a temp directory for loading
            with TemporaryDirectory() as td:
                cert_file = Path(td, "cert")
                key_file = Path(td, "key")
                write_cert_to_file(host_cert, str(cert_file))
                write_key_to_file(host_key, str(key_file))
                ctx.load_cert_chain(str(cert_file), str(key_file), args.ca_cert_key_pass)
        return ctx

    @staticmethod
    def __create_client_context(config: CoreConfig, ca_cert: Certificate) -> SSLContext:
        # noinspection PyTypeChecker
        ctx = create_default_context(purpose=Purpose.SERVER_AUTH)
        if config.args.ca_cert:
            ctx.load_verify_locations(cafile=config.args.ca_cert)
        else:
            ca_bytes = cert_to_bytes(ca_cert).decode("utf-8")
            ctx.load_verify_locations(cadata=ca_bytes)
        return ctx


class CertificateHandlerNoCA(CertificateHandler):
    """
    This certificate handler is using a CA that is outside the control of resoto core.
    As a consequence it is not possible to create new certificates or to sign CSRs.
    This functionality is only available in CertificateHandlerWithCA.

    All certificates can be provided via command line.
    If one part is missing, all of them get ignored.

    If not provided via command line, the CA certificate is retrieved from the CA.
    """

    @staticmethod
    def lookup(config: CoreConfig, temp_dir: Path) -> CertificateHandlerNoCA:
        args = config.args

        # if we get a ca certificate from the command line, use it
        if args.ca_cert and args.ca_cert_key and args.cert and args.cert_key:
            ca_cert = load_cert_from_file(args.ca_cert_cert)
            host_key = load_key_from_file(args.ca_cert_key, args.ca_cert_key_pass)
            host_cert = load_cert_from_file(args.ca_cert_cert)
            log.info(f"Using CA certificate from command line. fingerprint:{cert_fingerprint(ca_cert)}")
            return CertificateHandlerNoCA(config, ca_cert, host_key, host_cert, temp_dir)

        # otherwise, retrieve the CA certificate from the CA
        assert config.args.ca_url is not None, "CA URL must be set! Use --ca-url <url> to set it."
        cfg = config.api.host_certificate
        tls_data = TLSData(
            common_name=cfg.common_name,
            san_dns_names=get_local_hostnames(
                include_loopback=cfg.include_loopback,
                san_ip_addresses=cfg.san_ip_addresses,
                san_dns_names=cfg.san_dns_names,
            ),
            san_ip_addresses=get_local_ip_addresses(
                include_loopback=cfg.include_loopback, san_ip_addresses=cfg.san_ip_addresses
            ),
            resotocore_uri=config.args.ca_url,
            tempdir=str(temp_dir),
            psk=config.args.psk,
        )
        tls_data.load()
        return CertificateHandlerNoCA(config, tls_data.ca_cert, tls_data.key, tls_data.cert, temp_dir)


class CertificateHandlerWithCA(CertificateHandler):
    """
    This certificate handler is implementing the CA.
    It can create new certificates and sign CSRs.
    """

    def __init__(
        self,
        config: CoreConfig,
        ca_key: RSAPrivateKey,
        ca_cert: Certificate,
        host_key: RSAPrivateKey,
        host_cert: Certificate,
        temp_dir: Path,
    ) -> None:
        super().__init__(config, ca_cert, host_key, host_cert, temp_dir)
        self._ca_key = ca_key

    def create_key_and_cert(
        self, common_name: str, dns_names: List[str], ip_addresses: List[str], days_valid: int
    ) -> Tuple[RSAPrivateKey, Certificate]:
        key = gen_rsa_key()
        csr = gen_csr(
            key,
            include_loopback=False,
            common_name=common_name,
            san_dns_names=dns_names,
            san_ip_addresses=ip_addresses,
            discover_local_dns_names=False,
            discover_local_ip_addresses=False,
        )
        cert = sign_csr(csr, self._ca_key, self._ca_cert, days_valid)
        return key, cert

    def sign(
        self,
        csr_or_bytes: Union[CertificateSigningRequest, bytes],
        days_valid: int = 365,
        server_auth: bool = True,
        client_auth: bool = True,
        key_usage: Optional[Dict[str, bool]] = None,
    ) -> Tuple[Certificate, str]:
        csr = load_csr_from_bytes(csr_or_bytes) if isinstance(csr_or_bytes, bytes) else csr_or_bytes
        certificate = sign_csr(csr, self._ca_key, self._ca_cert, days_valid, server_auth, client_auth, key_usage)
        return certificate, cert_fingerprint(certificate)

    @staticmethod
    def _create_host_certificate(
        cfg: CertificateConfig, ca_key: RSAPrivateKey, ca_cert: Certificate
    ) -> Tuple[RSAPrivateKey, Certificate]:
        key = gen_rsa_key()
        host_names = get_local_hostnames(
            include_loopback=cfg.include_loopback,
            san_ip_addresses=cfg.san_ip_addresses,
            san_dns_names=cfg.san_dns_names,
        )
        host_ips = get_local_ip_addresses(include_loopback=cfg.include_loopback, san_ip_addresses=cfg.san_ip_addresses)
        log.info(f'Create host certificate for hostnames:{", ".join(host_names)} and ips:{", ".join(host_ips)}')
        csr = gen_csr(
            key,
            common_name=cfg.common_name,
            san_dns_names=list(host_names),
            san_ip_addresses=list(host_ips),
            include_loopback=cfg.include_loopback,
        )
        cert = sign_csr(csr, ca_key, ca_cert)
        return key, cert

    @staticmethod
    def lookup(config: CoreConfig, db: StandardDatabase, temp_dir: Path) -> CertificateHandlerWithCA:
        args = config.args
        # if we get a ca certificate from the command line, use it
        if args.ca_cert and args.ca_cert_key:
            ca_key = load_key_from_file(args.ca_cert_key, args.ca_cert_key_pass)
            ca_cert = load_cert_from_file(args.ca_cert_cert)
            if args.cert and args.cert_key:
                host_key = load_key_from_file(args.ca_cert_key, args.ca_cert_key_pass)
                host_cert = load_cert_from_file(args.ca_cert_cert)
            else:
                host_key, host_cert = CertificateHandlerWithCA._create_host_certificate(
                    config.api.host_certificate, ca_key, ca_cert
                )
            log.info(f"Using CA certificate from command line. fingerprint:{cert_fingerprint(ca_cert)}")
            return CertificateHandlerWithCA(config, ca_key, ca_cert, host_key, host_cert, temp_dir)

        # otherwise, load from database or create it
        sd = db.collection("system_data")
        maybe_ca: Optional[Json] = sd.get("ca")  # type: ignore
        if maybe_ca and isinstance(maybe_ca.get("key"), str) and isinstance(maybe_ca.get("certificate"), str):
            log.debug("Found existing certificate in data store.")
            key = load_key_from_bytes(maybe_ca["key"].encode("utf-8"))
            certificate = load_cert_from_bytes(maybe_ca["certificate"].encode("utf-8"))
            log.info(f"Using CA certificate from database. fingerprint:{cert_fingerprint(certificate)}")
            host_key, host_cert = CertificateHandlerWithCA._create_host_certificate(
                config.api.host_certificate, key, certificate
            )
            return CertificateHandlerWithCA(config, key, certificate, host_key, host_cert, temp_dir)
        else:
            wo = "with" if args.ca_cert_key_pass else "without"
            key, certificate = bootstrap_ca()
            log.info(
                f"No ca certificate found - create new one {wo} passphrase. fingerprint:{cert_fingerprint(certificate)}"
            )
            key_string = key_to_bytes(key, args.ca_cert_key_pass).decode("utf-8")
            certificate_string = cert_to_bytes(certificate).decode("utf-8")
            sd.insert({"_key": "ca", "key": key_string, "certificate": certificate_string})
            host_key, host_cert = CertificateHandlerWithCA._create_host_certificate(
                config.api.host_certificate, key, certificate
            )
            return CertificateHandlerWithCA(config, key, certificate, host_key, host_cert, temp_dir)



from attrs import define
from datetime import datetime
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import ArangoEntityDb
from resotocore.model.graph_access import DeferredEdge
from resotocore.ids import TaskId
from typing import List, cast
import logging

from resotocore.model.typed_model import from_js
from resotocore.types import Json
from resotocore.ids import GraphName


@define
class DeferredOuterEdges:
    id: str
    change_id: str
    task_id: TaskId
    created_at: datetime  # update the corresponding TTL index when changing this name
    graph: GraphName
    edges: List[DeferredEdge]


TWO_HOURS = 7200


log = logging.getLogger(__name__)


class DeferredOuterEdgeDb(ArangoEntityDb[str, DeferredOuterEdges]):
    async def all_for_task(self, task_id: TaskId) -> List[DeferredOuterEdges]:
        result = []
        async with await self.db.aql_cursor(
            f"FOR e IN `{self.collection_name}` FILTER e.task_id == @task_id RETURN e", bind_vars={"task_id": task_id}
        ) as cursor:
            async for doc in cursor:
                edges = from_js(doc, DeferredOuterEdges)
                result.append(edges)
        return result

    async def delete_for_task(self, task_id: TaskId) -> None:
        async with await self.db.aql_cursor(
            f"FOR e IN `{self.collection_name}` FILTER e.task_id == @task_id REMOVE e IN `{self.collection_name}`",
            bind_vars={"task_id": task_id},
        ) as cursor:
            async for _ in cursor:
                pass

    async def create_update_schema(self) -> None:
        await super().create_update_schema()
        ttl_index_name = "deferred_edges_expiration_index"
        collection = self.db.collection(self.collection_name)
        if ttl_index_name not in {idx["name"] for idx in cast(List[Json], collection.indexes())}:
            log.info(f"Add index {ttl_index_name} on {collection.name}")
            collection.add_ttl_index(["created_at"], TWO_HOURS, "deferred_edges_expiration_index")


def deferred_outer_edge_db(db: AsyncArangoDB, collection: str) -> DeferredOuterEdgeDb:
    return DeferredOuterEdgeDb(db, collection, DeferredOuterEdges, lambda k: k.id)



from typing import Dict

from pytest import fixture

from resotocore.cli.cli import CLIService
from resotocore.config import ConfigEntity
from resotocore.db.model import QueryModel
from resotocore.ids import ConfigId, GraphName
from resotocore.query.model import P, Query
from resotocore.report import BenchmarkConfigRoot, CheckConfigRoot, BenchmarkResult
from resotocore.report.inspector_service import (
    InspectorService,
    check_id,
    benchmark_id,
    InspectorConfigService,
    InspectorFileService,
)
from resotocore.report.report_config import (
    config_model,
    ReportCheckCollectionConfig,
    BenchmarkConfig,
)
from resotocore.types import Json
from resotocore.util import partition_by


@fixture
async def inspector_service_with_test_benchmark(
    cli: CLIService, inspection_check_collection: Json, benchmark: Json
) -> InspectorService:
    service = InspectorConfigService(cli)
    await service.config_handler.put_config(ConfigEntity(check_id("test"), inspection_check_collection))
    await service.config_handler.put_config(ConfigEntity(benchmark_id("test"), benchmark))
    return service


@fixture
async def inspector_file_service(cli: CLIService) -> InspectorService:
    return InspectorFileService(cli)


@fixture
def inspection_check_collection() -> Json:
    return dict(
        report_check=dict(
            provider="test",
            service="test",
            checks=[
                dict(
                    title="search",
                    name="search",
                    result_kind="foo",
                    categories=["test"],
                    severity="critical",
                    risk="",
                    # we use a query with a template here
                    detect={"resoto": "is({{foo_kind}})"},
                    default_values={"foo_kind": "foo"},
                    remediation=dict(text="", url="", action={}),
                ),
                dict(
                    title="cmd",
                    name="cmd",
                    result_kind="foo",
                    categories=["test"],
                    severity="critical",
                    detect={"resoto_cmd": "search is(foo) | jq --no-rewrite ."},
                    risk="",
                    remediation=dict(text="", url="", action={}),
                ),
            ],
        )
    )


@fixture
def benchmark() -> Json:
    return dict(
        report_benchmark=dict(
            title="test_benchmark",
            description="test_benchmark",
            id="test",
            framework="test",
            version="1.0",
            checks=["test_test_search", "test_test_cmd"],
            clouds=["test"],
        )
    )


async def test_config_model() -> None:
    models = config_model()
    assert len(models) == 6


async def test_list_inspect_checks(inspector_service: InspectorService) -> None:
    # list all available checks
    all_checks = {i.id: i for i in await inspector_service.list_checks()}
    assert len(all_checks) >= 30

    # use different filter options. The more filter are used, fewer results are returned
    filter_options = dict(
        provider="aws",
        service="ec2",
        category="security",
        kind="aws_ec2_instance",
        check_ids=["aws_ec2_internet_facing_with_instance_profile"],
    )
    last_len = len(all_checks)
    for options in range(1, len(filter_options)):
        args = dict(list(filter_options.items())[0:options])
        matching_checks = [i for i in await inspector_service.list_checks(**args)]  # type: ignore
        assert len(matching_checks) > 0
        assert len(matching_checks) <= last_len
        last_len = len(matching_checks)
    assert last_len < len(all_checks)


async def test_perform_benchmark(inspector_service_with_test_benchmark: InspectorService) -> None:
    def assert_result(results: Dict[str, BenchmarkResult]) -> None:
        result = results["test"]
        assert result.checks[0].number_of_resources_failing == 10
        assert result.checks[1].number_of_resources_failing == 10
        filtered = result.filter_result(filter_failed=True)
        assert filtered.checks[0].number_of_resources_failing == 10
        assert len(filtered.checks[0].resources_failing_by_account["sub_root"]) == 10
        assert filtered.checks[1].number_of_resources_failing == 10
        assert len(filtered.checks[1].resources_failing_by_account["sub_root"]) == 10
        passing, failing = result.passing_failing_checks_for_account("sub_root")
        assert len(passing) == 0
        assert len(failing) == 2
        passing, failing = result.passing_failing_checks_for_account("does_not_exist")
        assert len(passing) == 2
        assert len(failing) == 0

    inspector = inspector_service_with_test_benchmark
    graph_name = GraphName(inspector.cli.env["graph"])
    performed = await inspector.perform_benchmarks(graph_name, ["test"], sync_security_section=True)
    assert_result(performed)

    # make sure the result is persisted as part of the node
    async def count_vulnerable() -> int:
        db = inspector.db_access.get_graph_db(graph_name)
        model = await inspector.model_handler.load_model(graph_name)
        all_vunerable = Query.by(P("security.has_issues") == True)  # noqa
        async with await db.search_list(QueryModel(all_vunerable, model), with_count=True) as cursor:
            return cursor.count()  # type: ignore

    assert await count_vulnerable() == 10

    # loading the result from the db should give the same information
    loaded = await inspector.load_benchmarks(graph_name, ["test"])
    assert_result(loaded)


async def test_benchmark_node_result(inspector_service_with_test_benchmark: InspectorService) -> None:
    inspector = inspector_service_with_test_benchmark
    results = await inspector.perform_benchmarks(inspector.cli.env["graph"], ["test"])
    result = results["test"]
    node_edge_list = result.to_graph()
    nodes, edges = partition_by(lambda x: x["type"] == "node", node_edge_list)
    assert len(node_edge_list) == 5  # 3 nodes + 2 edges
    assert len(nodes) == 3
    assert len(edges) == 2
    for edge in edges:
        assert edge["from"] == result.node_id


async def test_predefined_checks(inspector_service: InspectorService) -> None:
    checks = ReportCheckCollectionConfig.from_files()
    assert len(checks) > 0
    for name, check in checks.items():
        assert (await inspector_service.validate_check_collection_config({CheckConfigRoot: check})) is None


async def test_predefined_benchmarks(inspector_service: InspectorService) -> None:
    benchmarks = BenchmarkConfig.from_files()
    assert len(benchmarks) > 0
    for name, check in benchmarks.items():
        config = {BenchmarkConfigRoot: check}
        cfg_id = ConfigId(name)
        assert (await inspector_service.validate_benchmark_config(cfg_id, config)) is None
        benchmark = BenchmarkConfig.from_config(ConfigEntity(cfg_id, config))
        assert benchmark.clouds == ["aws"]


async def test_list_failing(inspector_service_with_test_benchmark: InspectorService) -> None:
    inspector = inspector_service_with_test_benchmark
    graph = inspector.cli.env["graph"]
    search_res = [r async for r in await inspector.list_failing_resources(graph, "test_test_search")]
    assert len(search_res) == 10
    cmd_res = [r async for r in await inspector.list_failing_resources(graph, "test_test_cmd")]
    assert len(cmd_res) == 10
    search_res_account = [r async for r in await inspector.list_failing_resources(graph, "test_test_search", ["n/a"])]
    assert len(search_res_account) == 0
    cmd_res_account = [r async for r in await inspector.list_failing_resources(graph, "test_test_cmd", ["n/a"])]
    assert len(cmd_res_account) == 0


async def test_file_inspector(inspector_file_service: InspectorService) -> None:
    assert len(await inspector_file_service.list_benchmarks()) >= 1
    assert len(await inspector_file_service.filter_checks()) >= 100
    aws_cis = await inspector_file_service.benchmark("aws_cis_1_5")
    assert aws_cis is not None



from typing import Callable, Awaitable, Union
from datetime import timedelta
import asyncio


async def eventually(
    predicate: Callable[[], Union[bool, Awaitable[bool]]],
    timeout: timedelta = timedelta(seconds=5),
    interval: timedelta = timedelta(seconds=0.1),
) -> None:
    async def wait_for_condition() -> None:
        predicate_result = predicate()
        if isinstance(predicate_result, Awaitable):
            predicate_result = await predicate_result
        while not predicate_result:
            await asyncio.sleep(interval.total_seconds())

    await asyncio.wait_for(wait_for_condition(), timeout.total_seconds())



from datetime import timedelta
from typing import Tuple, List, Any

import pytest

from resotocore.dependencies import Dependencies, TenantDependencyCache, TenantDependencies
from resotocore.service import Service
from resotocore.system_start import parse_args
from resotocore.types import JsonElement


def test_parse_override() -> None:
    def parse(args: str) -> List[Tuple[str, JsonElement]]:
        return parse_args(args.split()).config_override  # type: ignore

    assert parse(f"--override a=foo") == [("a", "foo")]
    assert parse(f"--override a=foo,bla") == [("a", ["foo", "bla"])]
    assert parse(f"--override a=foo,bla b=a,b,c") == [("a", ["foo", "bla"]), ("b", ["a", "b", "c"])]
    assert parse(f'--override a="value,with,comma,in,quotes"') == [("a", "value,with,comma,in,quotes")]
    assert parse(f'--override a=some,value,"with,comma"') == [("a", ["some", "value", "with,comma"])]


class ExampleService(Service):
    def __init__(self, name: str) -> None:
        self.name = name
        self.started = False

    async def start(self) -> Any:
        self.started = True

    async def stop(self) -> None:
        self.started = False


@pytest.mark.asyncio
async def test_nested_dependencies() -> None:
    deps = Dependencies(a=ExampleService("a"), b=ExampleService("b"))
    async with deps.tenant_dependencies(a=ExampleService("na"), c=ExampleService("c")) as td:
        assert len(deps.services) == 2  # manages a and b
        assert len(td.services) == 2  # manages a and c
        # Deps has a and b, but not c
        a = deps.service("a", ExampleService)
        b = deps.service("b", ExampleService)
        assert a is not None
        assert b is not None
        assert a.started is False
        assert b.started is False
        assert a.name == "a"
        with pytest.raises(KeyError):
            deps.service("c", ExampleService)
        # Tenant deps have a, b and c; a is another one than in deps
        a = td.service("a", ExampleService)
        b = td.service("b", ExampleService)
        c = td.service("c", ExampleService)
        assert a is not None
        assert b is not None
        assert c is not None
        assert a.name == "na"
        assert a.started is True
        assert b.started is False
        assert c.started is True


@pytest.mark.asyncio
async def test_dependency_cache() -> None:
    time = 1
    created = 0
    deps = Dependencies(a=ExampleService("a"))

    async def tenant_deps() -> TenantDependencies:
        nonlocal created
        created += 1
        return deps.tenant_dependencies(b=ExampleService("b"))

    async with TenantDependencyCache(timedelta(seconds=1), timedelta(days=1), lambda: time) as cache:
        # getting the value from the cache will create a new value
        assert created == 0
        td_1 = await cache.get("a", tenant_deps)
        assert td_1 is not None
        assert td_1.service("b", ExampleService).started is True
        assert created == 1
        # getting it again will return the same value
        assert await cache.get("a", tenant_deps) is td_1
        assert created == 1
        # calling _expire will not remove the value from the cache
        await cache._expire()
        assert await cache.get("a", tenant_deps) is td_1
        assert created == 1
        # move in time forwards. The value will be removed from the cache and a new one is created
        time = 3
        await cache._expire()
        td_3 = await cache.get("a", tenant_deps)
        assert created == 2
        assert td_3 is not None and td_3 is not td_1
        # all services from td_1 are stopped, all services from td_3 are started
        assert td_1.service("b", ExampleService).started is False
        assert td_3.service("b", ExampleService).started is True
    # when the cache is stopped, all started services are stopped
    assert td_3.service("b", ExampleService).started is False



from __future__ import annotations
from enum import Enum
from typing import Set, Dict, Any

from attr import define


class Permission(Enum):
    read = "read"  # can read all resource data
    write = "write"  # can change all resource data
    admin = "admin"  # can change configuration


@define
class Role:
    name: str
    permissions: Set[Permission]

    def has_permission(self, permission: Permission) -> bool:
        return permission in self.permissions


PredefineRoles = {
    r.name: r
    for r in [
        Role("admin", {Permission.admin, Permission.read, Permission.write}),
        Role("readwrite", {Permission.read, Permission.write}),
        Role("readonly", {Permission.read}),
        Role("service", {Permission.admin, Permission.read, Permission.write}),
    ]
}
AllowedRoleNames = set(PredefineRoles.keys()) - {"service"}


@define
class AuthorizedUser:
    email: str
    roles: Set[str]
    permissions: Set[Permission]
    is_user: bool

    def has_permission(self, required_permissions: Set[Permission]) -> bool:
        return required_permissions.issubset(self.permissions)

    @staticmethod
    def from_jwt(jwt: Dict[str, Any]) -> AuthorizedUser:
        def permissions_for_role(role_name: str) -> Set[Permission]:
            if role_name not in PredefineRoles:
                return set()
            return PredefineRoles[role_name].permissions

        if "email" in jwt:  # This is a user token
            email = jwt["email"]
            roles = set(jwt["roles"].split(","))
            is_user = True
        else:  # This is a service token
            email = "service@resoto.com"
            roles = {"service"}
            is_user = False
        return AuthorizedUser(
            email=email,
            roles=roles,
            permissions={perm for rn in roles for perm in permissions_for_role(rn)},
            is_user=is_user,
        )



import asyncio
from typing import Any

import pytest

# noinspection PyUnresolvedReferences
from aiohttp.pytest_plugin import aiohttp_client
from aiohttp.test_utils import TestClient
from aiohttp.web import Application, Request, Response
from multidict import CIMultiDict
from pytest import mark

from resotocore.dependencies import TenantDependencies
from resotocore.user.model import Permission
from resotocore.web.auth import AuthHandler
from resotolib.jwt import encode_jwt


@pytest.fixture
async def loop() -> Any:
    return asyncio.get_running_loop()


@pytest.fixture
async def app_with_auth(auth_handler: AuthHandler) -> Application:
    async def hello(_: Request, deps: TenantDependencies) -> Response:
        return Response(text="Hello, world")

    app = Application(middlewares=[auth_handler.middleware()])
    app.router.add_get("/", auth_handler.allow_with(hello))
    app.router.add_get("/with_read", auth_handler.allow_with(hello, Permission.read))
    app.router.add_get("/with_write", auth_handler.allow_with(hello, Permission.write))
    return app


@mark.asyncio
async def test_correct_psk(aiohttp_client: Any, app_with_auth: Application) -> None:
    client: TestClient = await aiohttp_client(app_with_auth)
    jwt = encode_jwt({"foo": "bla"}, "test")
    resp = await client.get("/", headers=CIMultiDict({"Authorization": f"Bearer {jwt}"}))
    assert resp.status == 200


@mark.asyncio
async def test_wrong_psk(aiohttp_client: Any, app_with_auth: Application) -> None:
    client: TestClient = await aiohttp_client(app_with_auth)
    jwt = encode_jwt({"foo": "bla"}, "wrong!")
    resp = await client.get("/", headers=CIMultiDict({"Authorization": f"Bearer {jwt}"}))
    assert resp.status == 401


@mark.asyncio
async def test_no_psk(aiohttp_client: Any, app_with_auth: Application) -> None:
    client: TestClient = await aiohttp_client(app_with_auth)
    resp = await client.get("/")
    assert resp.status == 401


@mark.asyncio
async def test_permission(aiohttp_client: Any, app_with_auth: Application) -> None:
    client: TestClient = await aiohttp_client(app_with_auth)
    jwt = encode_jwt({"email": "batman@gotham", "roles": "readonly"}, "test")
    headers = CIMultiDict({"Authorization": f"Bearer {jwt}"})
    assert (await client.get("/", headers=headers)).status == 200
    assert (await client.get("/with_read", headers=headers)).status == 200
    failing = await client.get("/with_write", headers=headers)
    assert failing.status == 403
    assert await failing.text() == "Not allowed to perform this operation. Missing permission: write"



import asyncio
import logging
import platform
import sys
import traceback
import warnings
from argparse import Namespace
from asyncio import Queue
from contextlib import suppress
from datetime import timedelta, datetime, timezone
from functools import partial
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import AsyncIterator, List, Union, cast

from aiohttp.web_app import Application
from attrs import evolve
from urllib3.exceptions import HTTPWarning

from resotocore import version
from resotocore.action_handlers.merge_outer_edge_handler import MergeOuterEdgesHandler
from resotocore.analytics import CoreEvent, NoEventSender
from resotocore.analytics.posthog import PostHogEventSender
from resotocore.analytics.recurrent_events import emit_recurrent_events
from resotocore.cli.cli import CLIService
from resotocore.cli.command import alias_names, all_commands
from resotocore.config.config_handler_service import ConfigHandlerService
from resotocore.config.config_override_service import ConfigOverrideService, model_from_db, override_config_for_startup
from resotocore.config.core_config_handler import CoreConfigHandler
from resotocore.core_config import (
    config_from_db,
    RunConfig,
    inside_docker,
    inside_kubernetes,
    helm_installation,
    ResotoCoreConfigId,
    parse_config,
    CoreConfig,
)
from resotocore.db import SystemData
from resotocore.db.db_access import DbAccess
from resotocore.db.system_data_db import EphemeralJwtSigningKey
from resotocore.dependencies import Dependencies, ServiceNames, TenantDependencies
from resotocore.dependencies import (
    TenantDependencyProvider,
    FromRequestTenantDependencyProvider,
    DirectTenantDependencyProvider,
)
from resotocore.error import RestartService
from resotocore.graph_manager.graph_manager import GraphManager
from resotocore.infra_apps.local_runtime import LocalResotocoreAppRuntime
from resotocore.infra_apps.package_manager import PackageManager
from resotocore.message_bus import MessageBus
from resotocore.model.db_updater import GraphMerger
from resotocore.model.model_handler import ModelHandlerDB, ModelHandlerFromCodeAndDB
from resotocore.model.typed_model import to_json, class_fqn
from resotocore.query.template_expander_service import TemplateExpanderService
from resotocore.report.inspector_service import InspectorConfigService, InspectorFileService
from resotocore.system_start import db_access, setup_process, parse_args, system_info, reconfigure_logging
from resotocore.task.scheduler import APScheduler, NoScheduler
from resotocore.task.subscribers import SubscriptionHandlerService
from resotocore.task.task_handler import TaskHandlerService
from resotocore.user.user_management import UserManagementService
from resotocore.util import shutdown_process, utc
from resotocore.web.api import Api
from resotocore.web.certificate_handler import CertificateHandlerWithCA, CertificateHandlerNoCA
from resotocore.worker_task_queue import WorkerTaskQueue
from resotolib.asynchronous.web import runner

log = logging.getLogger("resotocore")


def main() -> None:
    """
    Application entrypoint - no arguments are allowed.
    """
    try:
        run(sys.argv[1:])
        log.info("Process finished.")
    except (KeyboardInterrupt, SystemExit):
        log.info("Stopping resoto graph core.")
        shutdown_process(0)
    except Exception as ex:
        if "--debug" in sys.argv:
            print(traceback.format_exc())
        print(f"resotocore stopped. Reason {class_fqn(ex)}: {ex}", file=sys.stderr)
        shutdown_process(1)


def run(arguments: List[str]) -> None:
    """
    Run application. When this method returns, the process is done.
    :param arguments: the arguments provided to this process.
                 Note: this method is used in tests to specify arbitrary arguments.
    """
    args = parse_args(arguments)
    setup_process(args)

    # after setup, logging is possible
    info = system_info()
    log.info(
        f"Starting up version={info.version} on system with cpus={info.cpus}, "
        f"available_mem={info.mem_available}, total_mem={info.mem_total}"
    )

    # The loop is here to restart the process in case of RestartService exceptions.
    while True:
        try:
            run_process(args)
            break  # This line should never be reached. In case it does, break the loop.
        except RestartService as ex:
            message = f"Restarting Service. Reason: {ex.reason}"
            line = "-" * len(message)
            print(f"\n{line}\n{message}\n{line}\n")


def run_process(args: Namespace) -> None:
    with TemporaryDirectory() as temp_name:
        temp = Path(temp_name)
        if args.multi_tenant_setup:
            deps = Dependencies(system_info=system_info())
            deps.add(ServiceNames.temp_dir, temp)
            config = deps.add(ServiceNames.config, parse_config(args, {}, lambda: None))
            # jwt_signing_keys are not required for multi-tenant setup.
            deps.add(ServiceNames.jwt_signing_key_holder, EphemeralJwtSigningKey())
            cert_handler_no_ca = deps.add(ServiceNames.cert_handler, CertificateHandlerNoCA.lookup(config, temp))
            verify: Union[bool, str] = False if args.graphdb_no_ssl_verify else str(cert_handler_no_ca.ca_bundle)
            deps.add(ServiceNames.config, evolve(config, run=RunConfig(temp, verify)))
            deps.add(ServiceNames.system_data, SystemData("multi-tenant", datetime(2023, 9, 1, tzinfo=timezone.utc), 1))
            deps.add(
                ServiceNames.event_sender,
                PostHogEventSender(deps.system_data) if config.runtime.usage_metrics else NoEventSender(),
            )

            provider: TenantDependencyProvider = deps.add(
                ServiceNames.tenant_dependency_provider, FromRequestTenantDependencyProvider(deps)
            )
            created = False
        else:
            with warnings.catch_warnings():  # ignore ssl errors during setup
                deps = TenantDependencies(system_info=system_info())
                deps.add(ServiceNames.temp_dir, temp)
                warnings.simplefilter("ignore", HTTPWarning)
                # wait here for an initial connection to the database before we continue. blocking!
                created, system_data, sdb = DbAccess.connect(args, timedelta(seconds=120), verify=False)
                deps.add(ServiceNames.system_data, system_data)
                # only to be used for CoreConfig creation
                core_config_override_service = asyncio.run(override_config_for_startup(args.config_override_path))
                config = config_from_db(
                    args, sdb, lambda: core_config_override_service.get_override(ResotoCoreConfigId)
                )
                cert_handler = deps.add(ServiceNames.cert_handler, CertificateHandlerWithCA.lookup(config, sdb, temp))
                verify = False if args.graphdb_no_ssl_verify else str(cert_handler.ca_bundle)
                deps.add(ServiceNames.config, evolve(config, run=RunConfig(temp, verify)))
                # in case of tls: connect again with the correct certificate settings
                use_tls = args.graphdb_server.startswith("https://")
                db = DbAccess.connect(args, timedelta(seconds=30), verify=verify)[2] if use_tls else sdb
                deps.add(ServiceNames.system_database, db)
                event_sender = deps.add(
                    ServiceNames.event_sender,
                    PostHogEventSender(deps.system_data) if config.runtime.usage_metrics else NoEventSender(),
                )
                dba = deps.add(ServiceNames.db_access, db_access(config, sdb, event_sender))
                deps.add(ServiceNames.jwt_signing_key_holder, dba.system_data_db)
                provider = deps.add(ServiceNames.tenant_dependency_provider, DirectTenantDependencyProvider(deps))

        with_config(config, deps, provider, created)


async def direct_tenant(deps: TenantDependencies) -> None:
    config = deps.config
    event_sender = deps.event_sender
    db = deps.service(ServiceNames.db_access, DbAccess)
    message_bus = deps.add(ServiceNames.message_bus, MessageBus())
    scheduler = deps.add(ServiceNames.scheduler, APScheduler() if not config.args.no_scheduling else NoScheduler())
    model = deps.add(ServiceNames.model_handler, ModelHandlerDB(db, config.runtime.plantuml_server))
    worker_task_queue = deps.add(ServiceNames.worker_task_queue, WorkerTaskQueue())
    # a "real" config override deps.add, unlike the one used for core config
    config_override_service = deps.add(
        ServiceNames.config_override,
        ConfigOverrideService(config.args.config_override_path, partial(model_from_db, db.configs_model_db)),
    )
    config_handler = deps.add(
        ServiceNames.config_handler,
        ConfigHandlerService(
            db.config_entity_db,
            db.config_validation_entity_db,
            db.configs_model_db,
            worker_task_queue,
            message_bus,
            event_sender,
            config,
            config_override_service,
        ),
    )
    deps.add(ServiceNames.user_management, UserManagementService(db, config_handler, event_sender))
    default_env = {"graph": config.cli.default_graph, "section": config.cli.default_section}
    cli = deps.add(ServiceNames.cli, CLIService(deps, all_commands(deps), default_env, alias_names()))
    deps.add(ServiceNames.template_expander, TemplateExpanderService(db.template_entity_db, cli))
    inspector = deps.add(ServiceNames.inspector, InspectorConfigService(cli))
    subscriptions = deps.add(
        ServiceNames.subscription_handler, SubscriptionHandlerService(db.subscribers_db, message_bus)
    )
    core_config_handler = deps.add(
        ServiceNames.core_config_handler,
        CoreConfigHandler(config, message_bus, worker_task_queue, config_handler, event_sender, inspector),
    )
    deps.add(ServiceNames.infra_apps_runtime, LocalResotocoreAppRuntime(cli))
    deps.add(
        ServiceNames.infra_apps_package_manager,
        PackageManager(
            db.package_entity_db, config_handler, cli.register_infra_app_alias, cli.unregister_infra_app_alias
        ),
    )
    graph_merger = deps.add(ServiceNames.graph_merger, GraphMerger(model, event_sender, config, message_bus))
    task_handler = deps.add(
        ServiceNames.task_handler,
        TaskHandlerService(
            db.running_task_db,
            db.job_db,
            message_bus,
            event_sender,
            subscriptions,
            graph_merger,
            scheduler,
            cli,
            config,
        ),
    )
    deps.add(ServiceNames.graph_manager, GraphManager(db, config, core_config_handler, task_handler))
    deps.add(
        ServiceNames.merge_outer_edges_handler,
        MergeOuterEdgesHandler(message_bus, subscriptions, task_handler, db, model),
    )
    deps.add(
        ServiceNames.event_emitter_periodic,
        emit_recurrent_events(
            event_sender,
            model,
            subscriptions,
            worker_task_queue,
            message_bus,
            timedelta(hours=1),
            timedelta(hours=1),
        ),
    )
    # queue must be created inside an async function!
    deps.add(ServiceNames.forked_tasks, Queue())


async def multi_tenancy(deps: Dependencies) -> None:
    deps.add(ServiceNames.message_bus, MessageBus())
    deps.add(ServiceNames.forked_tasks, Queue())
    InspectorFileService.on_startup()
    ModelHandlerFromCodeAndDB.on_startup()


def with_config(
    config: CoreConfig, deps: Dependencies, tenant_dep_provider: TenantDependencyProvider, created: bool
) -> None:
    reconfigure_logging(config)  # based on the config, logging might have changed
    # only lg the editable config - to not log any passwords
    log.debug(f"Starting with config: {config.editable}")
    api = deps.add("api", Api(deps, tenant_dep_provider))

    async def on_start() -> None:
        # fill all dependencies
        if config.multi_tenant_setup:
            await multi_tenancy(deps)
        else:
            await direct_tenant(cast(TenantDependencies, deps))
        # start all dependencies
        await deps.start()
        # Send initial creation event
        if created:
            docker = inside_docker()
            kubernetes = inside_kubernetes()
            helm = helm_installation()
            await deps.event_sender.core_event(
                CoreEvent.SystemInstalled,
                {
                    "docker_install": docker,
                    "k8s_install": kubernetes,
                    "helm_install": helm,
                    "pip_install": not (docker or kubernetes or helm),
                },
            )

        await deps.event_sender.core_event(
            CoreEvent.SystemStarted,
            {
                "version": version(),
                "created_at": to_json(deps.system_data.created_at),
                "system": platform.system(),
                "platform": platform.platform(),
                "inside_docker": deps.system_info.inside_docker,
            },
            cpu_count=deps.system_info.cpus,
            mem_total=deps.system_info.mem_total,
            mem_available=deps.system_info.mem_available,
        )

    async def on_stop() -> None:
        duration = utc() - deps.system_info.started_at
        await deps.event_sender.core_event(CoreEvent.SystemStopped, total_seconds=int(duration.total_seconds()))
        await deps.stop()

    async def async_initializer() -> Application:
        async def clean_all_tasks() -> None:
            log.info("Clean up all running tasks.")
            for task in asyncio.all_tasks():
                with suppress(asyncio.CancelledError):
                    if not task.done() or not task.cancelled():
                        task.cancel()
                    log.debug(f"Wait for task: {task}")
                    await task

        async def on_start_stop(_: Application) -> AsyncIterator[None]:
            await on_start()
            log.info("Initialization done. Starting API.")
            yield
            log.info("Shutdown initiated. Stop all tasks.")
            await on_stop()
            await clean_all_tasks()

        api.app.cleanup_ctx.append(on_start_stop)
        return api.app

    runner.run_app(
        async_initializer(),
        api.stop,
        host=config.api.web_hosts,
        https_port=config.api.https_port,
        http_port=config.api.http_port,
        default_port=8900,
        ssl_context=deps.cert_handler.host_context,
    )


if __name__ == "__main__":
    main()



from datetime import datetime, timezone

import pytest
from pytest import fixture
from typing import List

from resotocore.db.deferredouteredgedb import DeferredOuterEdges, DeferredOuterEdgeDb
from resotocore.ids import TaskId, GraphName, NodeId
from resotocore.model.graph_access import DeferredEdge, ByNodeId, EdgeTypes


@fixture
def instances() -> List[DeferredOuterEdges]:
    return [
        DeferredOuterEdges(
            id="1",
            change_id="c1",
            task_id=TaskId("t1"),
            created_at=datetime(2021, 1, 1, tzinfo=timezone.utc),
            graph=GraphName("test"),
            edges=[DeferredEdge(ByNodeId(NodeId("e1")), ByNodeId(NodeId("e2")), EdgeTypes.default)],
        ),
        DeferredOuterEdges(
            id="2",
            change_id="c1",
            task_id=TaskId("t1"),
            created_at=datetime(2021, 1, 1, tzinfo=timezone.utc),
            graph=GraphName("test"),
            edges=[DeferredEdge(ByNodeId(NodeId("e2")), ByNodeId(NodeId("e3")), EdgeTypes.default)],
        ),
        DeferredOuterEdges(
            id="3",
            change_id="c2",
            task_id=TaskId("t2"),
            created_at=datetime(2021, 1, 1, tzinfo=timezone.utc),
            graph=GraphName("test"),
            edges=[DeferredEdge(ByNodeId(NodeId("e2")), ByNodeId(NodeId("e3")), EdgeTypes.default)],
        ),
    ]


@pytest.mark.asyncio
async def test_all_by_task_id(
    pending_deferred_edge_db: DeferredOuterEdgeDb, instances: List[DeferredOuterEdges]
) -> None:
    await pending_deferred_edge_db.update_many(instances)
    assert len(await pending_deferred_edge_db.all_for_task(TaskId("t1"))) == 2
    assert len(await pending_deferred_edge_db.all_for_task(TaskId("t2"))) == 1


@pytest.mark.asyncio
async def test_remove_by_task_id(
    pending_deferred_edge_db: DeferredOuterEdgeDb, instances: List[DeferredOuterEdges]
) -> None:
    await pending_deferred_edge_db.wipe()
    # insert all
    await pending_deferred_edge_db.update_many(instances)
    # delete t1
    await pending_deferred_edge_db.delete_for_task(TaskId("t1"))
    assert len(await pending_deferred_edge_db.all_for_task(TaskId("t1"))) == 0
    assert len(await pending_deferred_edge_db.all_for_task(TaskId("t2"))) == 1
    # delete t2
    await pending_deferred_edge_db.delete_for_task(TaskId("t2"))
    assert len([n async for n in pending_deferred_edge_db.all()]) == 0
